#include "Common.cl"




DECLARE_DYNAMIC_ARRAY( clPairedBlob, true );
DECLARE_DYNAMIC_ARRAY( clBlobPerson, true );



    
    

clCapsule2 GetFloorCapsule(const clCapsule2* BlobCapsule,const clCamera* Camera,const clPersonEstimateParams* PersonParams)
{
	float2 acenter = clCapsule2_GetCenter( BlobCapsule );

	//	calc the capsule
	float2 Floor2 = clCapsule2_GetCenterBottom( BlobCapsule );
	float4 Floor3 = clCamera_GetPosAtHeight( Camera, Floor2, 0.f, false );
	clCapsule2 Capsulea = clCamera_GetEstimatedPerson2( Camera, Floor3, *PersonParams );
	return Capsulea;
}





    

kernel void CalcBlobPairings(	global const clCapsule2*	FrameBlobCapsules, 
								global clPairedBlob*		PairedCapsules, 
								volatile global int*		PairedCapsuleCount,
								const int					PairedCapsuleMax,
								const clCamera				Camera,
								clPersonEstimateParams		PersonParams
								)
{
    
    //	gr: this struct is all zeroes atm (bug)
	PersonParams = _clPersonEstimateParams();
	clBlobPairParams BlobPairParams = _clBlobPairParams();

	
	int a = get_global_id(0);
	int FrameBlobCount = get_global_size(0);

	clCapsule2 BlobCapsulea = FrameBlobCapsules[a];

	//	get the capsule at the floor
	clCapsule2 PersonCapsulea = GetFloorCapsule( &BlobCapsulea, &Camera, &PersonParams );
	float2 PersonFloora = clCapsule2_GetCenterBottom( &PersonCapsulea );
	float2 PersonCeilinga = clCapsule2_GetCenterTop( &PersonCapsulea );

	TArray_clPairedBlob PairedArray = { PairedCapsules, PairedCapsuleCount, PairedCapsuleMax };

	//	iterate through following blobs to see if any can be paired up
	for ( int b=a+1;	b<FrameBlobCount;	b++ )
	{
		//	get capsule for blobb
		clCapsule2 BlobCapsuleb = FrameBlobCapsules[b];
		
		clCapsule2 PersonCapsule = PersonCapsulea;
		clCapsule2 CompareCapsule = BlobCapsuleb;

		//	compare lowest blob as base
		float2 Floorb = clCapsule2_GetCenterBottom( &BlobCapsuleb );
		if ( Floorb.y > PersonFloora.y )
		{
			CompareCapsule = BlobCapsulea;
			PersonCapsule = GetFloorCapsule( &BlobCapsuleb, &Camera, &PersonParams );
		}
		
		//	simple horizontal test first (get center-line distance and test if any overlap)
		float2 DistanceNormal = (float2)(0,0);
		if ( !clCapsule2_GetHorizontalOverlap( &DistanceNormal.x, &PersonCapsule, &CompareCapsule ) )
			continue;
		
		if ( !clCapsule2_GetVerticalOverlap( &DistanceNormal.y, &PersonCapsule, &CompareCapsule ) )
			continue;
		//	doesn't overlap enough
		//if ( DistanceNormal.x > 1.0f )
		//	continue;
		if ( fabs(DistanceNormal.y) > BlobPairParams.mMaxDistanceNormal.y )
			continue;

		//	do horizontal re-align
		//if ( BlobPairParams.mRealignPairCapsule )
		if ( true )
		{
			float NewLeft = ( clCapsule2_GetLeft(&PersonCapsule) + clCapsule2_GetLeft(&CompareCapsule) ) / 2.f;
			clCapsule2_SetLeft( &PersonCapsule, NewLeft );
		}

		//	push result
		clPairedBlob PairedBlob = { PersonCapsule, a, b };
        
		if ( !PushArray_clPairedBlob( PairedArray, &PairedBlob ) )
		{
			//	log items we couldnt output
			return;
		}
	   
    }
 

}


//	Go through the blob persons and find all the blob capsules that fit inside our person
//	and add to the list of blobs that make up that person
bool clBlobPerson_ContainsBlobIndex(const clBlobPerson* BlobPerson,int BlobIndex)
{
	for ( int i=0;	i<BlobPerson->mBlobCount;	i++ )
		if ( BlobPerson->mBlobIndex[i] == BlobIndex )
			return true;
	return false;
}

kernel void FindBlobPersonPairings(	global clBlobPerson* BlobPersons, 
										global const clCapsule2* BlobCapsules, 
										const int	BlobCapsulesCount,
										global const clCamera* pCamera,
										global const clPersonEstimateParams* pPersonParams,
										global const clBlobPairParams* pBlobPairParams
									)
{
	const clCamera Camera = *pCamera;
	const clPersonEstimateParams PersonParams = *pPersonParams;

	int a = get_global_id(0);
	int FrameBlobCount = get_global_size(0);
	global clBlobPerson* pBlobPerson = &BlobPersons[a];
	clBlobPerson BlobPerson = *pBlobPerson;
	clCapsule2 BlobCapsulea = BlobCapsules[ BlobPerson.mBlobIndex[0] ];
	clCapsule2 PersonCapsulea = BlobPerson.mCapsule;

	//	get the capsule at the floor
	float2 PersonFloora = clCapsule2_GetCenterBottom( &PersonCapsulea );
	float2 PersonCeilinga = clCapsule2_GetCenterTop( &PersonCapsulea );

	//	iterate through following blobs to see if any can be paired up
	for ( int b=0;	b<BlobCapsulesCount;	b++ )
	{
		//	already in person
		if ( clBlobPerson_ContainsBlobIndex( &BlobPerson, b ) )
			continue;
		
		//	output full
		if ( BlobPerson.mBlobCount >= sizeofarray(BlobPerson.mBlobIndex) )
			break;

		//	get capsule for blobb
		clCapsule2 BlobCapsuleb = BlobCapsules[b];
		
		clCapsule2 PersonCapsule = PersonCapsulea;
		clCapsule2 CompareCapsule = BlobCapsuleb;

		//	compare lowest blob as base
		float2 Floorb = clCapsule2_GetCenterBottom( &BlobCapsuleb );
		if ( Floorb.y > PersonFloora.y )
		{
			CompareCapsule = BlobCapsulea;
			PersonCapsule = GetFloorCapsule( &BlobCapsuleb, &Camera, &PersonParams );
		}
		
		//	simple horizontal test first (get center-line distance and test if any overlap)
		float2 DistanceNormal = (float2)(0,0);
		if ( !clCapsule2_GetHorizontalOverlap( &DistanceNormal.x, &PersonCapsule, &CompareCapsule ) )
			continue;
		
		if ( !clCapsule2_GetVerticalOverlap( &DistanceNormal.y, &PersonCapsule, &CompareCapsule ) )
			continue;
		//	doesn't overlap enough
		//if ( DistanceNormal.x > 1.0f )
		//	continue;
		if ( fabs(DistanceNormal.y) > pBlobPairParams->mMaxDistanceNormal.y )
			continue;

		//	add to list of blobs
		BlobPerson.mBlobIndex[ BlobPerson.mBlobCount++ ] = b;
	}

	//	re-write new blob person info
	*pBlobPerson = BlobPerson;
}


								
kernel void SplitBlobs(	global const clCapsule2* FrameBlobCapsules, 
							global clPairedBlob* SplitCapsules, 
							volatile global int* SplitCapsuleCount,
							const int			SplitCapsuleMax,
							const clCamera Camera,
							clPersonEstimateParams PersonParams
							)
{
	//	gr: this struct is all zeroes atm (bug)
	PersonParams = _clPersonEstimateParams();
	clBlobPairParams BlobPairParams = _clBlobPairParams();

	int a = get_global_id(0);
	int FrameBlobCount = get_global_size(0);

	clCapsule2 BlobCapsulea = FrameBlobCapsules[a];

	//	get the capsule at the floor
	clCapsule2 PersonCapsulea = GetFloorCapsule( &BlobCapsulea, &Camera, &PersonParams );
	float2 PersonFloora = clCapsule2_GetCenterBottom( &PersonCapsulea );
	float2 PersonCeilinga = clCapsule2_GetCenterTop( &PersonCapsulea );

	TArray_clPairedBlob PairedArray = { SplitCapsules, SplitCapsuleCount, SplitCapsuleMax };

	//	find multiple people in wide blobs...
	float PersonWidth = PersonCapsulea.mRadius * 2.f;
	float BlobLeft = clCapsule2_GetLeft(&BlobCapsulea);
	float BlobRight = clCapsule2_GetRight(&BlobCapsulea);
	while ( (BlobRight-BlobLeft) > PersonWidth * BlobPairParams.mWideBlobFactor )
	{
		//	first, get the outer ones, there's noise there, so place people at the edges
		clCapsule2 LeftCapsule = PersonCapsulea;
		clCapsule2_SetLeft( &LeftCapsule, BlobLeft );
		clPairedBlob LeftPairedBlob = { LeftCapsule, a, a };
		if ( !PushArray_clPairedBlob( PairedArray, &LeftPairedBlob ) )
		{
			//	log items we couldnt output
			return;
		}
		clCapsule2 RightCapsule = PersonCapsulea;
		clCapsule2_SetRight( &RightCapsule, BlobRight );
		clPairedBlob RightPairedBlob = { RightCapsule, a, a };
		if ( !PushArray_clPairedBlob( PairedArray, &RightPairedBlob ) )
		{
			//	log items we couldnt output
			return;
		}

		//	condense the blob and then next iteration will fill that
		BlobRight -= PersonWidth;
		BlobLeft += PersonWidth;
		
		//	see if we can fit 1 more in (rather than 2)
	}
}



kernel void GetBlobPersons(	global const clCapsule2* BlobCapsules, 
								global clBlobPerson* BlobPersons, 
								volatile global int* BlobPersonsCount, 
								const int BlobPersonsMax,
								global const clCamera* pCamera,
								global const clPersonEstimateParams* pPersonParams,
								global const clBlobPairParams* pBlobPairParams
								)
{
	const clCamera Camera = *pCamera;
	const clPersonEstimateParams PersonParams = *pPersonParams;
	
	const int BlobCapsuleCount = get_global_size(0);
	int a = get_global_id(0);
	clCapsule2 BlobCapsule = BlobCapsules[a];

	float2 acenter = clCapsule2_GetCenter( &BlobCapsule );

	//	calc the capsule
	float2 Floor2 = clCapsule2_GetCenterBottom( &BlobCapsule );
	float4 Floor3 = clCamera_GetPosAtHeight( &Camera, Floor2, 0.f, false );
	
	//	generate and store the center person
	clBlobPerson BlobPerson = _clBlobPerson();
	BlobPerson.mCapsule = clCamera_GetEstimatedPerson2( &Camera, Floor3, PersonParams );
	BlobPerson.mBlobIndex[BlobPerson.mBlobCount++] = a;
	TArray_clBlobPerson BlobPersonArray = { BlobPersons, BlobPersonsCount, BlobPersonsMax };

	if ( !PushArray_clBlobPerson( BlobPersonArray, &BlobPerson ) )
		return;	

	//	find multiple people in very wide blobs...
	float PersonWidth = BlobPerson.mCapsule.mRadius * 2.f;
	float BlobLeft = clCapsule2_GetLeft(&BlobCapsule);
	float BlobRight = clCapsule2_GetRight(&BlobCapsule);
	float WideFactor = pBlobPairParams->mWideBlobFactor;
	while ( (BlobRight-BlobLeft) > PersonWidth * WideFactor )
	{
		//	first, get the outer ones, there's noise there, so place people at the edges
		
		clCapsule2_SetLeft( &BlobPerson.mCapsule, BlobLeft );
		if ( !PushArray_clBlobPerson( BlobPersonArray, &BlobPerson ) )
			return;

		clCapsule2_SetRight( &BlobPerson.mCapsule, BlobRight );
		if ( !PushArray_clBlobPerson( BlobPersonArray, &BlobPerson ) )
			return;

		//	condense the blob and then next iteration will fill that
		BlobRight -= PersonWidth;
		BlobLeft += PersonWidth;
	}


}


